@@ -10,6 +10,7 @@
  */
 #include <inttypes.h>
 
+#include "node.h"
 #include "fsck.h"
 #include "xattr.h"
 #ifdef HAVE_ATTR_XATTR_H
@@ -194,7 +195,7 @@ void ssa_dump(struct f2fs_sb_info *sbi, int start_ssa, int end_ssa)
 {
 	struct f2fs_summary_block *sum_blk;
 	char buf[BUF_SZ];
-	int segno, i, ret;
+	int segno, type, i, ret;
 	int fd;
 
 	fd = open("dump_ssa", O_CREAT|O_WRONLY|O_TRUNC, 0666);
@@ -207,10 +208,10 @@ void ssa_dump(struct f2fs_sb_info *sbi, int start_ssa, int end_ssa)
 	ASSERT(ret >= 0);
 
 	for (segno = start_ssa; segno < end_ssa; segno++) {
-		sum_blk = get_sum_block(sbi, segno, &ret);
+		sum_blk = get_sum_block(sbi, segno, &type);
 
 		memset(buf, 0, BUF_SZ);
-		switch (ret) {
+		switch (type) {
 		case SEG_TYPE_CUR_NODE:
 			snprintf(buf, BUF_SZ, "\n\nsegno: %x, Current Node\n", segno);
 			break;
@@ -239,8 +240,8 @@ void ssa_dump(struct f2fs_sb_info *sbi, int start_ssa, int end_ssa)
 			ret = write(fd, buf, strlen(buf));
 			ASSERT(ret >= 0);
 		}
-		if (ret == SEG_TYPE_NODE || ret == SEG_TYPE_DATA ||
-					ret == SEG_TYPE_MAX)
+		if (type == SEG_TYPE_NODE || type == SEG_TYPE_DATA ||
+					type == SEG_TYPE_MAX)
 			free(sum_blk);
 	}
 	close(fd);
@@ -474,6 +475,18 @@ dump:
 	}
 }
 
+static bool is_sit_bitmap_set(struct f2fs_sb_info *sbi, u32 blk_addr)
+{
+	struct seg_entry *se;
+	u32 offset;
+
+	se = get_seg_entry(sbi, GET_SEGNO(sbi, blk_addr));
+	offset = OFFSET_IN_SEG(sbi, blk_addr);
+
+	return f2fs_test_bit(offset,
+			(const char *)se->cur_valid_map) != 0;
+}
+
 void dump_node(struct f2fs_sb_info *sbi, nid_t nid, int force)
 {
 	struct node_info ni;
@@ -491,15 +504,18 @@ void dump_node(struct f2fs_sb_info *sbi, nid_t nid, int force)
 
 	if (ni.blk_addr == 0x0)
 		MSG(force, "Invalid nat entry\n\n");
+	else if (!is_sit_bitmap_set(sbi, ni.blk_addr))
+		MSG(force, "Invalid node blk addr\n\n");
 
 	DBG(1, "node_blk.footer.ino [0x%x]\n", le32_to_cpu(node_blk->footer.ino));
 	DBG(1, "node_blk.footer.nid [0x%x]\n", le32_to_cpu(node_blk->footer.nid));
 
 	if (le32_to_cpu(node_blk->footer.ino) == ni.ino &&
-			le32_to_cpu(node_blk->footer.nid) == ni.nid &&
-			ni.ino == ni.nid) {
+			le32_to_cpu(node_blk->footer.nid) == ni.nid) {
 		print_node_info(sbi, node_blk, force);
-		dump_file(sbi, &ni, node_blk, force);
+
+		if (ni.ino == ni.nid)
+			dump_file(sbi, &ni, node_blk, force);
 	} else {
 		print_node_info(sbi, node_blk, force);
 		MSG(force, "Invalid (i)node block\n\n");
@@ -580,14 +596,104 @@ static void dump_node_offset(u32 blk_addr)
 	free(node_blk);
 }
 
+static int has_dirent(u32 blk_addr, int is_inline, int *enc_name)
+{
+	struct f2fs_node *node_blk;
+	int ret, is_dentry = 0;
+
+	node_blk = calloc(BLOCK_SZ, 1);
+	ASSERT(node_blk);
+
+	ret = dev_read_block(node_blk, blk_addr);
+	ASSERT(ret >= 0);
+
+	if (IS_INODE(node_blk) && S_ISDIR(le16_to_cpu(node_blk->i.i_mode)))
+		is_dentry = 1;
+
+	if (is_inline && !(node_blk->i.i_inline & F2FS_INLINE_DENTRY))
+		is_dentry = 0;
+
+	*enc_name = file_is_encrypt(&node_blk->i);
+
+	free(node_blk);
+
+	return is_dentry;
+}
+
+static void dump_dirent(u32 blk_addr, int is_inline, int enc_name)
+{
+	struct f2fs_dentry_ptr d;
+	void *inline_dentry, *blk;
+	int ret, i = 0;
+
+	blk = calloc(BLOCK_SZ, 1);
+	ASSERT(blk);
+
+	ret = dev_read_block(blk, blk_addr);
+	ASSERT(ret >= 0);
+
+	if (is_inline) {
+		inline_dentry = inline_data_addr((struct f2fs_node *)blk);
+		make_dentry_ptr(&d, blk, inline_dentry, 2);
+	} else {
+		make_dentry_ptr(&d, NULL, blk, 1);
+	}
+
+	DBG(1, "%sDentry block:\n", is_inline ? "Inline " : "");
+
+	while (i < d.max) {
+		struct f2fs_dir_entry *de;
+		unsigned char en[F2FS_NAME_LEN + 1];
+		u16 en_len, name_len;
+		int enc;
+
+		if (!test_bit_le(i, d.bitmap)) {
+			i++;
+			continue;
+		}
+
+		de = &d.dentry[i];
+
+		if (!de->name_len) {
+			i++;
+			continue;
+		}
+
+		name_len = le16_to_cpu(de->name_len);
+		enc = enc_name;
+
+		if (de->file_type == F2FS_FT_DIR) {
+			if ((d.filename[i][0] == '.' && name_len == 1) ||
+				(d.filename[i][0] == '.' &&
+				d.filename[i][1] == '.' && name_len == 2)) {
+				enc = 0;
+			}
+		}
+
+		en_len = convert_encrypted_name(d.filename[i],
+				le16_to_cpu(de->name_len), en, enc);
+		en[en_len] = '\0';
+
+		DBG(1, "bitmap pos[0x%x] name[%s] len[0x%x] hash[0x%x] ino[0x%x] type[0x%x]\n",
+				i, en,
+				le16_to_cpu(de->name_len),
+				le32_to_cpu(de->hash_code),
+				le32_to_cpu(de->ino),
+				de->file_type);
+
+		i += GET_DENTRY_SLOTS(le16_to_cpu(de->name_len));
+	}
+
+	free(blk);
+}
+
 int dump_info_from_blkaddr(struct f2fs_sb_info *sbi, u32 blk_addr)
 {
 	nid_t nid;
 	int type;
 	struct f2fs_summary sum_entry;
 	struct node_info ni, ino_ni;
-	struct seg_entry *se;
-	u32 offset;
+	int enc_name;
 	int ret = 0;
 
 	MSG(0, "\n== Dump data from block address ==\n\n");
@@ -619,12 +725,8 @@ int dump_info_from_blkaddr(struct f2fs_sb_info *sbi, u32 blk_addr)
 		return ret;
 	}
 
-	se = get_seg_entry(sbi, GET_SEGNO(sbi, blk_addr));
-	offset = OFFSET_IN_SEG(sbi, blk_addr);
-
-	if (f2fs_test_bit(offset, (const char *)se->cur_valid_map) == 0) {
+	if (!is_sit_bitmap_set(sbi, blk_addr))
 		MSG(0, "\nblkaddr is not valid\n");
-	}
 
 	type = get_sum_entry(sbi, blk_addr, &sum_entry);
 	nid = le32_to_cpu(sum_entry.nid);
@@ -666,12 +768,18 @@ int dump_info_from_blkaddr(struct f2fs_sb_info *sbi, u32 blk_addr)
 		dump_node_from_blkaddr(sbi, ino_ni.blk_addr);
 		dump_data_offset(ni.blk_addr,
 			le16_to_cpu(sum_entry.ofs_in_node));
+
+		if (has_dirent(ino_ni.blk_addr, 0, &enc_name))
+			dump_dirent(blk_addr, 0, enc_name);
 	} else {
 		MSG(0, "FS Userdata Area: Node block from 0x%x\n", blk_addr);
 		if (ni.ino == ni.nid) {
 			MSG(0, " - Inode block       : id = 0x%x from 0x%x\n",
 					ni.ino, ino_ni.blk_addr);
 			dump_node_from_blkaddr(sbi, ino_ni.blk_addr);
+
+			if (has_dirent(ino_ni.blk_addr, 1, &enc_name))
+				dump_dirent(blk_addr, 1, enc_name);
 		} else {
 			MSG(0, " - Node block        : id = 0x%x from 0x%x\n",
 					nid, ni.blk_addr);
